/*  The C equivalent of TUdataLSq2D,  It consists of a TCL interface over   */
/*  a straight C subroutine                                                 */
/*                                                                          */
/*  There are 10 OBJV elements.  These are:                                 */
/*     OBJV[0]:  Routine name                                               */
/*     OBJV[1]:  Array of X data (X)                                        */
/*     OBJV[2]:  Array of Y data (Y)                                        */
/*     OBJV[3]:  Array of Z data (Z)                                        */
/*     OBJV[4]:  Elements In Data Arrays (N)                                */
/*     OBJV[5]:  Order Of Fit (nOrder)                                      */
/*     OBJV[6]:  Statistical Method to apply (Mode)                         */
/*     OBJV[7]:  Solution Coefficients (Coef)                               */
/*     OBJV[8]:  Varience Data To Be Produced (Vr)                          */
/*     OBJV[9]:  Variations of Z data (SigZ)                                */

#include <tcl.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "ByteDefs.h"
#include "TutilAnsi.h"

int DataLSq2DCmd (ClientData cD, Tcl_Interp *interp, 
                                int objc, Tcl_Obj *CONST *objv) 
{
   void *memPtr = NULL;
   ReaL_8 *X, *Y, *Z, *SigZ, VarV;
   ReaL_8 Coef[10];
   int N, nOrder, Mode, Vr;
   ByTe_4 nSigZ;
   ByTe_1 OuT[40];

   ByTe_4 Bytes, rV;

   Tcl_Obj *objPtr;

/* Make sure that the minimum number of IO parameters are present          */

   if (objc < 8 ) {
     Tcl_WrongNumArgs(interp, 1, objv, "Usage: TUdataLSq2D X Y Z ... ");
     return TCL_ERROR;
   }

/* Get the constant input parameters                                        */

   Tcl_GetIntFromObj (interp, objv[4], &N);
   Tcl_GetIntFromObj (interp, objv[5], &nOrder);
   Tcl_GetIntFromObj (interp, objv[6], &Mode);
   if (objc >= 9)
      Tcl_GetIntFromObj (interp, objv[8], &Vr);

/* See if there is a SigZ array present                                     */

   if (Mode > 1) { Mode = 0; } 
   if (Mode != 1) {
      nSigZ = 0;
      SigZ = NULL;
   } else { nSigZ = N; }

/* Get the input arrays                                                     */

   Bytes = sizeof(ReaL_8) * (3 * N  + nSigZ);
   if ((memPtr = malloc (Bytes)) == NULL)
      return TCL_ERROR;
   X = (ReaL_8 *)memPtr;
   Y = (ReaL_8 *)memPtr + N;
   Z = (ReaL_8 *)memPtr + 2 * N;

   if (TclArrayToC (interp, objv[1], (void *)X, N, 0, 'D') == 0) {
      free (memPtr);
      return TCL_OK;
   }
   if (TclArrayToC (interp, objv[2], (void *)Y, N, 0, 'D') == 0) {
      free (memPtr);
      return TCL_OK;
   }
   if (TclArrayToC (interp, objv[3], (void *)Z, N, 0, 'D') == 0) {
      free (memPtr);
      return TCL_OK;
   }

/* If present get the SigZ vectors                                     */

   if (nSigZ > 0) {
      SigZ = (ReaL_8 *)memPtr + 3 * N;
      if (TclArrayToC (interp, objv[9], (void *)SigZ, N, 0, 'D') == 0) {
         free (memPtr);
         return TCL_OK;
      }
   }

/* Call the  straight C array procedure                                   */

   X = (ReaL_8 *)memPtr;
   Y = (ReaL_8 *)memPtr + N;
   Z = (ReaL_8 *)memPtr + 2 * N;
   if (nSigZ > 0) { SigZ = (ReaL_8 *)memPtr + 3 * N; }

   rV = DataLSq2D_C (X, Y, Z, N, nOrder, Mode, Coef, Vr, SigZ, &VarV); 

   if (CArrayToTcl (interp, objv[7], (void *)Coef, rV, 0, 'D') != 1) {
      free (memPtr);
      return TCL_OK;
   }

/* Gather output - free data array - and gone                             */

   free (memPtr);

   if (Vr == 1 ) {
      sprintf (OuT, "%d %.8e", rV, VarV);
      objPtr = Tcl_GetObjResult(interp);
      Tcl_SetStringObj (objPtr, OuT, strlen(OuT));
   }
   return TCL_OK;

}

/*  This is a 2D Least Squares Code - straight C                           */

ByTe_4 DataLSq2D_C (ReaL_8 *X, ReaL_8 *Y, ReaL_8 *Z, ByTe_4 N, ByTe_4 nOrder,
                  ByTe_4 Mode, ReaL_8 *A,  ByTe_4 Var, ReaL_8 *SiGZ, ReaL_8 *V)
{
   void *memPtr;
   ReaL_8 *Mat, *SumV, Dif, *TmpXY, Dummy[2];
   ReaL_8 S[N], diS, pW;
   ByTe_4 I, J, K, L, M; 
   ByTe_4 RanK, Bytes, Base;
   ByTe_4 MEnd, EndXY, EndV, LenX, Len, End;
   ByTe_4 Start, StartX, Cnt;

/* Set up the statistics array                                             */

   switch (Mode) { 
      case 0: 
         for (I = 0; I < N; ++I) { S[I] = 1.0; }
      break;
      case 1: 
         for (I = 0; I < N; ++I) { S[I] = SiGZ[I]; }
      break;
      default: 
         pW = -Mode / 2.0;
         for (I = 0; I < N; ++I) {
            diS = X[I] * X[I] + Y[I] * Y[I]; 
            S[I] = (diS != 0.0 ) ? 1.0 / pow(diS, pW) : 1.0;
         }
      break;
   }

/* Set the rand of the needed matrix                                        */

   RanK = 1;
   for (I = 1; I <= nOrder; ++I) { RanK += I + 1; }

/* Zero Out the matrix and solution vector                                  */

   MEnd = RanK * RanK;
   Bytes = sizeof(ReaL_8) * (2 * RanK * RanK + RanK);
   if ((memPtr = malloc (Bytes)) == NULL)
     return -1;
   Mat = (ReaL_8 *)memPtr;
   SumV = (ReaL_8 *)memPtr + MEnd;
   TmpXY = (ReaL_8 *)memPtr + MEnd + RanK;

   for (I = 0; I < MEnd; ++I) { Mat[I] = 0.0; }
   for (I = 0; I < RanK; ++I) { SumV[I] = 0.0; }

   EndXY = 2 * nOrder;
   EndV = nOrder + 1;

   for (I = 0; I < N; ++I) {
      Base = 0;
      LenX = 1;
      TmpXY[0] = S[I];
      L = 1;
      for (J = 0; J < EndXY; ++J) {
         M = Base;
         for (K = 0; K < LenX; ++K) {
            TmpXY[L++] = TmpXY[M++] * X[I]; 
         }
         --M;
         TmpXY[L++] = TmpXY[M] * Y[I];
         Base += LenX;
         ++LenX;
      }

/*  Put together the matrix                                                 */

      for (J = 0; J < RanK; ++J) { Mat[J] += TmpXY[J]; }

      StartX = 1;
      LenX = 2;
      Cnt = 0;
      for ( ; J < MEnd; ++StartX ) {
         Len = LenX;
         Start = StartX;
         for (K = 0; K < EndV; ++K) { 
             M = Start;
             End = K + 1;
             for (L = 0; L < End; ++L) { Mat[J++] += TmpXY[M++]; }
             Start += Len;
             ++Len; 
         }
         ++Cnt;
         if ( Cnt == LenX ) {
            Cnt = 0;
            ++LenX;
         }
      }
         
      for (J = 0; J < RanK; ++J) { SumV[J] += TmpXY[J] * Z[I]; }
   }

   MatrixInv_C (RanK, Mat, 1, SumV);
   for (I = 0; I < RanK; ++I) { A[I] = SumV[I]; }

   *V = 0.0;
   if ( Var == 1) {
      for (I = 0; I < N; ++I) {
         Dif = Z[I] - PolyExp2D_C (X[I], Y[I], RanK, A, nOrder, 0, Dummy);
         *V += Dif * Dif;
      }
      *V =  sqrt(*V / (ReaL_8)N);
   } 

   free (memPtr);
   return (RanK);
}
